@@ -42,7 +42,6 @@
 #include <sys/sysinfo.h>
 #include <sys/stat.h>
 #include <sys/mman.h>
-#include <sys/signalfd.h>
 #ifdef CHECK_FORDEVPTS
 # include <sys/vfs.h>
 #endif
@@ -52,6 +51,12 @@
 #include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
+#ifdef USE_BLOGD
+# include <libblogger.h>
+#else
+# define bootlog(arg...)
+# define closeblog()
+#endif
 #include "makeboot.h"
 #include "proc.h"
 
@@ -197,6 +202,7 @@ void closeall(void)
   for (s = 0; s < par; s++)
     if (prgs[s].fd)
       close(prgs[s].fd);
+  closeblog();
 }
 
 void callsplash(int n, const char *path, char *action)
@@ -241,23 +247,16 @@ void callsplash(int n, const char *path, char *action)
       return;
     }
 
-  (void)sigemptyset(&nmask);
-  (void)sigaddset(&nmask, SIGINT);
-  (void)sigaddset(&nmask, SIGHUP);
-  (void)sigaddset(&nmask, SIGQUIT);
-  (void)sigaddset(&nmask, SIGSEGV);
-  (void)sigaddset(&nmask, SIGTERM);
-  (void)sigaddset(&nmask, SIGCHLD);
-  (void)sigaddset(&nmask, SIGTTIN);
+  (void)sigfillset(&nmask);
   sigprocmask(SIG_UNBLOCK, &nmask, NULL);
 
-  (void)signal(SIGINT,  SIG_DFL);
   (void)signal(SIGHUP,  SIG_DFL);
   (void)signal(SIGQUIT, SIG_DFL);
   (void)signal(SIGSEGV, SIG_DFL);
   (void)signal(SIGTERM, SIG_DFL);
   (void)signal(SIGCHLD, SIG_DFL);
   (void)signal(SIGTTIN, SIG_DFL);
+  (void)signal(SIGTTOU, SIG_DFL);
 
   TEMP_FAILURE_RETRY(dup2(2, 1));
   closeall();
@@ -456,29 +455,23 @@ void run(struct prg *p)
       return;
     }
 
-  (void)sigemptyset(&nmask);
-  (void)sigaddset(&nmask, SIGINT);
-  (void)sigaddset(&nmask, SIGHUP);
-  (void)sigaddset(&nmask, SIGQUIT);
-  (void)sigaddset(&nmask, SIGSEGV);
-  (void)sigaddset(&nmask, SIGTERM);
-  (void)sigaddset(&nmask, SIGCHLD);
-  (void)sigaddset(&nmask, SIGTTIN);
+  (void)sigfillset(&nmask);
   sigprocmask(SIG_UNBLOCK, &nmask, NULL);
 
-  (void)signal(SIGINT,  SIG_DFL);
   (void)signal(SIGHUP,  SIG_DFL);
   (void)signal(SIGQUIT, SIG_DFL);
   (void)signal(SIGSEGV, SIG_DFL);
   (void)signal(SIGTERM, SIG_DFL);
   (void)signal(SIGCHLD, SIG_DFL);
   (void)signal(SIGTTIN, SIG_DFL);
+  (void)signal(SIGTTOU, SIG_DFL);
 
   if (setpgid(0, 0))
     perror("setpgid");
 
   if (m && p->fd)
     {
+      sigset_t smask, omask;
       TEMP_FAILURE_RETRY(close(1));
       if (open(m, O_RDWR) != 1)
 	{
@@ -486,8 +479,12 @@ void run(struct prg *p)
 	  _exit(1);
 	}
       TEMP_FAILURE_RETRY(dup2(1, 2));
-      if (tcsetattr(1, TCSANOW, &tio))
+      sigemptyset(&smask);
+      sigaddset(&smask, SIGTTOU);
+      sigprocmask(SIG_BLOCK, &smask, &omask);
+      if (tcsetattr(1, TCSANOW, &tio) && errno != ENOTTY)
 	perror("tcsetattr");
+      sigprocmask(SIG_SETMASK, &omask, NULL);
       if (wzok)
 	ioctl(1, TIOCSWINSZ, &wz);
       putenv(sz.env_row);
@@ -501,7 +498,7 @@ void run(struct prg *p)
   closeall();
 
 #ifdef USE_PRELOAD
-  if (!strcmp(arg, "start")) 
+  if (arg && !strcmp(arg, "start")) 
     { 
       int s;
 
@@ -582,23 +579,17 @@ int run_single(const char *prg, const char *arg0, int spl)
     {
       sigset_t nmask;
 
-      (void)sigemptyset(&nmask);
-      (void)sigaddset(&nmask, SIGINT);
-      (void)sigaddset(&nmask, SIGHUP);
-      (void)sigaddset(&nmask, SIGQUIT);
-      (void)sigaddset(&nmask, SIGSEGV);
-      (void)sigaddset(&nmask, SIGTERM);
-      (void)sigaddset(&nmask, SIGCHLD);
-      (void)sigaddset(&nmask, SIGTTIN);
+      (void)sigfillset(&nmask);
       sigprocmask(SIG_UNBLOCK, &nmask, NULL);
 
-      (void)signal(SIGINT,  SIG_DFL);
+      (void)signal(SIGINT,  SIG_DFL);	/* Interactive init scripts may be interrupted */
       (void)signal(SIGHUP,  SIG_DFL);
       (void)signal(SIGQUIT, SIG_DFL);
       (void)signal(SIGSEGV, SIG_DFL);
       (void)signal(SIGTERM, SIG_DFL);
       (void)signal(SIGCHLD, SIG_DFL);
       (void)signal(SIGTTIN, SIG_DFL);
+      (void)signal(SIGTTOU, SIG_DFL);
 
       TEMP_FAILURE_RETRY(dup2(2, 1));
       closeall();
@@ -632,19 +623,20 @@ void do_forward(void)
 	{
 	  if (errno == EINTR)
 	    continue;
+#if defined(DEBUG) && (DEBUG > 0)
 	  perror("\n\rstartpar: forward read");
+#endif
 	  break;
 	}
       b = buf;
       while (r > 0)
 	{
 	  rr = write(1, b, r);
-	  if (rr == -1)
+	  if (rr < 0)
 	    {
 	      if (errno == EINTR)
 		continue;
 	      perror("\n\rstartpar: forward write");
-	      break;
 	      rr = r;
 	    }
 	  r -= rr;
@@ -730,9 +722,22 @@ void detach(struct prg *p, const int store)
     {
       if ((pid = fork()) == 0)
 	{
+	  sigset_t nmask;
+	  (void)sigfillset(&nmask);
+	  sigprocmask(SIG_UNBLOCK, &nmask, NULL);
+
+	  (void)signal(SIGHUP,  SIG_DFL);
+	  (void)signal(SIGQUIT, SIG_DFL);
+	  (void)signal(SIGSEGV, SIG_DFL);
+	  (void)signal(SIGTERM, SIG_DFL);
+	  (void)signal(SIGCHLD, SIG_DFL);
+	  (void)signal(SIGTTIN, SIG_DFL);
+	  (void)signal(SIGTTOU, SIG_DFL);
+
 	  TEMP_FAILURE_RETRY(dup2(p->fd, 0));
 	  TEMP_FAILURE_RETRY(dup2(2, 1));
 	  closeall();
+
 	  execlp(myname, myname, "-f", "--", p->name, NULL);
 	  do_forward();
 	}
@@ -824,6 +829,7 @@ int main(int argc, char **argv)
   int forw = 0;
   int c, i, num;
   int limit;
+  int notty = 0;
   int *resvec;
   struct timeval tv;
   struct prg *p;
@@ -833,10 +839,13 @@ int main(int argc, char **argv)
   char *splashopt = 0;
   sigset_t nmask, omask, smask;
 
+  detect_consoles();
+
   (void)sigemptyset(&nmask);
   (void)sigaddset(&nmask, SIGHUP);
   sigprocmask(SIG_UNBLOCK, &nmask, NULL);
 
+  (void)signal(SIGINT,  SIG_IGN);	/* Init scripts should not be interruptible */
   (void)signal(SIGCHLD, SIG_DFL);
   numcpu = sysconf(_SC_NPROCESSORS_ONLN);
   myname = argv[0];
@@ -1015,10 +1024,17 @@ int main(int argc, char **argv)
 
   if (tcgetattr(0, &tio))
     {
-      perror("tcgetattr");
+      if (errno != ENOTTY)
+	perror("tcgetattr");
       tcgetattr(2, &tio);
+      if (errno == ENOTTY)
+	notty = 1;
     }
   cfmakeraw(&tio);
+  tio.c_lflag &= ~ECHO;
+  tio.c_lflag |= ISIG;
+  tio.c_cc[VTIME] = 0;
+  tio.c_cc[VMIN]  = CMIN;
 
   if (ioctl(0, TIOCGWINSZ, &wz) == 0)
     wzok = 1;
@@ -1148,6 +1164,8 @@ int main(int argc, char **argv)
 		  if (!devpts)
 		    interactive_task = p;	/* no /dev/pts, treat as interactive */
 #endif
+		  if (notty)
+		    interactive_task = p;	/* no tty, treat as interactive */
 		  if (interactive_task)
 		    continue;			/* don't start this here */
 		  run(p);
